:root {
  font-family: system-ui, Avenir, Helvetica, Arial, sans-serif;
  line-height: 1.5;
  font-weight: 400;

  color-scheme: light dark;
  color: var(--color-primary);
  background-color: var(--color-surface-01);

  font-synthesis: none;
  text-rendering: optimizeLegibility;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}

#root {
  max-width: 900px;
  margin: 0 auto;
  padding: 2rem;
  text-align: center;
  width: 100%;
}

body {
  margin: 0;
  display: flex;
  place-items: center;
  min-height: 100vh;
}
